"""MQTT Hermes client base class"""
import asyncio
import io
import json
import logging
import queue
import subprocess
import threading
import typing
import wave
from concurrent.futures import CancelledError
from pathlib import Path

from .asr import AsrTrain
from .audioserver import AudioFrame, AudioSessionFrame, AudioSummary
from .base import Message
from .nlu import NluTrain

# -----------------------------------------------------------------------------

TopicArgs = typing.Mapping[str, typing.Any]
GeneratorType = typing.AsyncIterable[
    typing.Optional[typing.Union[Message, typing.Tuple[Message, TopicArgs]]]
]

# -----------------------------------------------------------------------------


class HermesClient:
    """Base class for Hermes MQTT clients"""

    def __init__(
        self,
        client_name: str,
        mqtt_client,
        site_ids: typing.Optional[typing.List[str]] = None,
        sample_rate: int = 16000,
        sample_width: int = 2,
        channels: int = 1,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
    ):
        # Internal logger
        self.client_name = client_name
        self.logger = logging.getLogger(client_name)

        # Paho MQTT client
        self.mqtt_client = mqtt_client
        self.mqtt_client.on_connect = self.mqtt_on_connect
        self.mqtt_client.on_disconnect = self.mqtt_on_disconnect
        self.mqtt_client.on_message = self.mqtt_on_message

        # Set when on_connect succeeds
        self.mqtt_connected_event: asyncio.Event = asyncio.Event()

        self.is_connected: bool = False
        self.subscribe_lock = threading.Lock()
        self.pending_mqtt_topics: typing.Set[str] = set()

        # Incoming message queue (async)
        self.in_queue: typing.Optional[asyncio.Queue] = None
        self.pre_queue: queue.Queue = queue.Queue()

        # Message types that are subscribed to
        self.subscribed_types: typing.Set[typing.Type[Message]] = set()
        self.subscribed_topics: typing.Set[str] = set()

        # Cache of all MQTT topics in case we get disconnected
        self.all_mqtt_topics: typing.Set[str] = set()

        # Set of valid site ids (empty for all)
        self.site_ids: typing.Set[str] = set(site_ids) if site_ids else set()
        self.site_id = "default" if not site_ids else site_ids[0]

        # Required audio format
        self.sample_rate = sample_rate
        self.sample_width = sample_width
        self.channels = channels

        self.loop: typing.Optional[asyncio.AbstractEventLoop] = loop

    # -------------------------------------------------------------------------
    # User Methods
    # -------------------------------------------------------------------------

    def subscribe(self, *message_types: typing.Type[Message], **topic_args):
        """Subscribe to one or more Hermes messages."""
        topics: typing.List[str] = []

        if self.site_ids:
            # Specific site ids
            for site_id in self.site_ids:
                for message_type in message_types:
                    topics.append(message_type.topic(site_id=site_id))
                    self.subscribed_types.add(message_type)
        else:
            # All site ids
            for message_type in message_types:
                topics.append(message_type.topic())
                self.subscribed_types.add(message_type)

        # Subscribe to all MQTT topics
        self.subscribe_topics(*topics)

    def subscribe_topics(self, *topics):
        """Subscribe to one or more MQTT topics."""
        with self.subscribe_lock:
            self.pending_mqtt_topics.update(topics)

            if self.is_connected:
                # Subscribe to all pending topics
                for topic in self.pending_mqtt_topics:
                    self.all_mqtt_topics.add(topic)

                    # Don't re-subscribe
                    if topic not in self.subscribed_topics:
                        self.mqtt_client.subscribe(topic)
                        self.subscribed_topics.add(topic)
                        self.logger.debug("Subscribed to %s", topic)

                self.pending_mqtt_topics.clear()

    async def on_message(
        self,
        message: Message,
        site_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        topic: typing.Optional[str] = None,
    ) -> GeneratorType:
        """Override to handle Hermes messages."""
        yield None

    async def on_message_blocking(
        self,
        message: Message,
        site_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        topic: typing.Optional[str] = None,
    ) -> GeneratorType:
        """Override to handle Hermes messages and block."""
        yield None

    async def on_raw_message(self, topic: str, payload: bytes):
        """Override to handle MQTT messages."""
        pass

    # -------------------------------------------------------------------------
    # MQTT Event Handlers
    # -------------------------------------------------------------------------

    def mqtt_on_connect(self, client, userdata, flags, rc):
        """Connected to MQTT broker."""
        try:
            self.is_connected = True
            self.logger.debug("Connected to MQTT broker")

            # Clear topic cache
            self.subscribed_topics.clear()

            # Re-subscribe to everything if previous disconnected
            self.pending_mqtt_topics.update(self.all_mqtt_topics)

            # Handle subscriptions
            self.subscribe()

            if self.loop:
                self.loop.call_soon_threadsafe(self.mqtt_connected_event.set)
        except Exception:
            self.logger.exception("on_connect")

    def mqtt_on_disconnect(self, client, userdata, flags, rc):
        """Automatically reconnect when disconnected."""
        try:
            self.logger.warning("Disconnected. Trying to reconnect...")

            # Automatically reconnect
            if self.loop:
                self.loop.call_soon_threadsafe(self.mqtt_connected_event.clear)

            self.is_connected = False

            self.mqtt_client.reconnect()
        except Exception:
            self.logger.exception("on_disconnect")

    def mqtt_on_message(self, client, userdata, msg):
        """Received message from MQTT broker."""
        try:
            # Handle message in event loop
            if self.loop and self.in_queue:
                self.loop.call_soon_threadsafe(self.in_queue.put_nowait, msg)
            else:
                # Save in pre-queue to be picked up later
                self.pre_queue.put(msg)
        except Exception:
            self.logger.exception("on_message")

    async def handle_messages_async(
        self, loop: typing.Optional[asyncio.AbstractEventLoop] = None
    ):
        """Handles MQTT messages in event loop."""
        self.loop = loop or self.loop or asyncio.get_running_loop()
        self.in_queue = asyncio.Queue()

        # Pull in messages from pre-queue
        while self.pre_queue.qsize() > 0:
            self.in_queue.put_nowait(self.pre_queue.get_nowait())

        # Main loop
        while True:
            try:
                mqtt_message = await self.in_queue.get()
                if mqtt_message is None:
                    break

                # Fire and forget
                asyncio.create_task(
                    self.on_raw_message(mqtt_message.topic, mqtt_message.payload)
                )

                # Check against all known message types
                for message, site_id, session_id in HermesClient.parse_mqtt_message(
                    mqtt_message.topic,
                    mqtt_message.payload,
                    self.subscribed_types,
                    logger=self.logger,
                ):

                    if not self.valid_site_id(site_id):
                        continue

                    # Log messages
                    if message.is_binary_payload():
                        # Class name + size
                        if not isinstance(message, (AudioFrame, AudioSessionFrame)):
                            self.logger.debug(
                                "<- %s(%s byte(s))",
                                message.__class__.__name__,
                                len(mqtt_message.payload),
                            )
                    elif isinstance(message, (AsrTrain, NluTrain)):
                        # Just class name
                        self.logger.debug("<- %s", message.__class__.__name__)
                    elif not isinstance(message, AudioSummary):
                        # Entire message
                        self.logger.debug("<- %s", message)

                    # Publish all responses (blocking)
                    await self.publish_all(
                        self.on_message_blocking(
                            message,
                            site_id=site_id,
                            session_id=session_id,
                            topic=mqtt_message.topic,
                        )
                    )

                    # Publish all responses (non-blocking)
                    asyncio.create_task(
                        self.publish_all(
                            self.on_message(
                                message,
                                site_id=site_id,
                                session_id=session_id,
                                topic=mqtt_message.topic,
                            )
                        )
                    )
            except KeyboardInterrupt:
                break
            except CancelledError:
                break
            except Exception:
                self.logger.exception("handle_messages_async")
                break

    @classmethod
    def parse_mqtt_message(
        cls,
        topic: str,
        payload: typing.Union[str, bytes],
        subscribed_types: typing.Iterable[typing.Type[Message]],
        logger=None,
    ) -> typing.Iterable[
        typing.Tuple[Message, typing.Optional[str], typing.Optional[str]]
    ]:
        """Deserialize MQTT message into Hermes object."""
        try:
            # Check against all known message types
            for message_type in subscribed_types:
                if message_type.is_topic(topic):
                    site_id: typing.Optional[str] = None

                    # Verify site id and parse
                    if message_type.is_binary_payload():
                        # Binary
                        if message_type.is_site_in_topic():
                            site_id = message_type.get_site_id(topic)

                        # Assume payload is only argument to constructor
                        message = message_type(payload)  # type: ignore
                    else:
                        # JSON
                        json_payload = json.loads(payload)
                        if message_type.is_site_in_topic():
                            site_id = message_type.get_site_id(topic)
                        else:
                            site_id = json_payload.get("siteId")

                        # Load from JSON
                        message = message_type.from_dict(json_payload)

                    session_id: typing.Optional[str] = None
                    if message_type.is_session_in_topic():
                        session_id = message_type.get_session_id(topic)

                    yield (message, site_id, session_id)

                    # Assume only one message type will match
                    break
        except Exception:
            if not logger:
                logger = logging

            logger.exception("parse_mqtt_message (topic=%s)", topic)

    # -------------------------------------------------------------------------
    # Publishing Messages
    # -------------------------------------------------------------------------

    def publish(self, message: Message, **topic_args):
        """Publish a Hermes message to MQTT."""
        try:
            topic = message.topic(**topic_args)
            payload = message.payload()

            if message.is_binary_payload():
                # Don't log audio frames
                if not isinstance(message, (AudioFrame, AudioSessionFrame)):
                    self.logger.debug(
                        "-> %s(%s byte(s))", message.__class__.__name__, len(payload)
                    )
            else:
                # Log most JSON messages
                if isinstance(message, (AsrTrain, NluTrain)):
                    # Just class name
                    self.logger.debug("-> %s", message.__class__.__name__)
                    self.logger.debug(
                        "Publishing %s bytes(s) to %s", len(payload), topic
                    )
                elif not isinstance(message, AudioSummary):
                    # Entire message
                    self.logger.debug("-> %s", message)
                    self.logger.debug(
                        "Publishing %s bytes(s) to %s", len(payload), topic
                    )

            self.mqtt_client.publish(topic, payload)
        except Exception:
            self.logger.exception(
                "publish (message=%s, topic_args=%s)",
                message.__class__.__name__,
                topic_args,
            )

    async def publish_all(self, async_generator: GeneratorType):
        """Enumerate all messages in an async generator publish them"""
        async for maybe_message in async_generator:
            if maybe_message is None:
                continue

            if isinstance(maybe_message, Message):
                self.publish(maybe_message)
            else:
                message, kwargs = maybe_message
                self.publish(message, **kwargs)

    # -------------------------------------------------------------------------
    # Utility Methods
    # -------------------------------------------------------------------------

    def valid_site_id(self, site_id: typing.Optional[str]):
        """True if site id is valid for this client."""
        if site_id and self.site_ids:
            return site_id in self.site_ids

        return True

    def convert_wav(
        self,
        wav_bytes: bytes,
        sample_rate: typing.Optional[int] = None,
        sample_width: typing.Optional[int] = None,
        channels: typing.Optional[int] = None,
    ) -> bytes:
        """Converts WAV data to required format with sox. Return raw audio."""
        if sample_rate is None:
            sample_rate = self.sample_rate

        if sample_width is None:
            sample_width = self.sample_width

        if channels is None:
            channels = self.channels

        return subprocess.run(
            [
                "sox",
                "-t",
                "wav",
                "-",
                "-r",
                str(sample_rate),
                "-e",
                "signed-integer",
                "-b",
                str(sample_width * 8),
                "-c",
                str(channels),
                "-t",
                "raw",
                "-",
            ],
            check=True,
            stdout=subprocess.PIPE,
            input=wav_bytes,
        ).stdout

    def maybe_convert_wav(
        self,
        wav_bytes: bytes,
        sample_rate: typing.Optional[int] = None,
        sample_width: typing.Optional[int] = None,
        channels: typing.Optional[int] = None,
    ) -> bytes:
        """Converts WAV data to required format if necessary. Returns raw audio."""
        if sample_rate is None:
            sample_rate = self.sample_rate

        if sample_width is None:
            sample_width = self.sample_width

        if channels is None:
            channels = self.channels

        with io.BytesIO(wav_bytes) as wav_io:
            with wave.open(wav_io, "rb") as wav_file:
                if (
                    (wav_file.getframerate() != sample_rate)
                    or (wav_file.getsampwidth() != sample_width)
                    or (wav_file.getnchannels() != channels)
                ):
                    # Return converted wav
                    return self.convert_wav(
                        wav_bytes,
                        sample_rate=sample_rate,
                        sample_width=sample_width,
                        channels=channels,
                    )

                # Return original audio
                return wav_file.readframes(wav_file.getnframes())

    def to_wav_bytes(
        self,
        audio_data: bytes,
        sample_rate: typing.Optional[int] = None,
        sample_width: typing.Optional[int] = None,
        channels: typing.Optional[int] = None,
    ) -> bytes:
        """Wrap raw audio data in WAV."""
        if sample_rate is None:
            sample_rate = self.sample_rate

        if sample_width is None:
            sample_width = self.sample_width

        if channels is None:
            channels = self.channels

        with io.BytesIO() as wav_buffer:
            wav_file: wave.Wave_write = wave.open(wav_buffer, mode="wb")
            with wav_file:
                wav_file.setframerate(sample_rate)
                wav_file.setsampwidth(sample_width)
                wav_file.setnchannels(channels)
                wav_file.writeframes(audio_data)

            return wav_buffer.getvalue()

    def reduce_noise(
        self, audio_data: bytes, noise_profile: Path, amount: float = 0.5
    ) -> bytes:
        """Reduce noise in raw audio using sox noise profile."""
        return subprocess.run(
            [
                "sox",
                "-r",
                str(self.sample_rate),
                "-e",
                "signed-integer",
                "-b",
                str(self.sample_width * 8),
                "-c",
                str(self.channels),
                "-t",
                "raw",
                "-",
                "-t",
                "raw",
                "-",
                "noisered",
                str(noise_profile),
                str(amount),
            ],
            check=True,
            stdout=subprocess.PIPE,
            input=audio_data,
        ).stdout



"""Messages for automated speech recognition."""
import re
import typing
from dataclasses import dataclass
from enum import Enum

from .base import Message
from .nlu import AsrToken


class AsrToggleReason(str, Enum):
    """Reason for ASR toggle on/off.

    Values
    ------
    UNKNOWN
        Overrides all other reasons

    DIALOGUE_SESSION
        Dialogue session is active

    PLAY_AUDIO
        Audio is currently playing

    TTS_SAY
        Text to speech system is currently speaking
    """

    UNKNOWN = ""
    DIALOGUE_SESSION = "dialogueSession"
    PLAY_AUDIO = "playAudio"
    TTS_SAY = "ttsSay"


@dataclass
class AsrToggleOn(Message):
    """Activate the ASR component.

    Attributes
    ----------
    site_id: str = "default"
        The id of the site where ASR should be turned on

    reason: AsrToggleReason = UNKNOWN
        Reason why ASR was toggled on
    """

    site_id: str = "default"

    # ------------
    # Rhasspy only
    # ------------

    reason: AsrToggleReason = AsrToggleReason.UNKNOWN

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/asr/toggleOn"


@dataclass
class AsrToggleOff(Message):
    """Deactivate the ASR component.

    Attributes
    ----------
    site_id: str = "default"
        The id of the site where ASR should be turned off

    reason: AsrToggleReason = UNKNOWN
        Reason why ASR was toggled off
    """

    site_id: str = "default"

    # ------------
    # Rhasspy only
    # ------------

    reason: AsrToggleReason = AsrToggleReason.UNKNOWN

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/asr/toggleOff"


@dataclass
class AsrStartListening(Message):
    """Tell the ASR component to start listening.

    Attributes
    ----------
    site_id: str = "default"
        The site that must be listened too

    session_id: Optional[str] = None
        An optional session id if there is a related session

    stop_on_silence: bool = True
        If true, ASR should automatically detect end of voice command

    send_audio_captured: bool = False
        If true, ASR emits asr/audioCaptured message with recorded audio

    wakeword_id: Optional[str] = None
        Optional id of wakeword used to activate ASR

    intent_filter: Optional[List[str]] = None
        A list of intent names to restrict the ASR on

    lang: Optional[str] = None
        Language of the incoming audio stream.
        Typically set in hotword detected or dialogue startSession messages.
    """

    site_id: str = "default"
    session_id: typing.Optional[str] = None
    lang: typing.Optional[str] = None

    # ------------
    # Rhasspy only
    # ------------

    stop_on_silence: bool = True
    send_audio_captured: bool = False
    wakeword_id: typing.Optional[str] = None
    intent_filter: typing.Optional[typing.List[str]] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/asr/startListening"


@dataclass
class AsrStopListening(Message):
    """Tell the ASR component to stop listening.

    Attributes
    ----------
    site_id: str = "default"
        The id of the site where the ASR should stop listening

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    site_id: str = "default"
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/asr/stopListening"


@dataclass
class AsrTextCaptured(Message):
    """Full ASR transcription results.

    Attributes
    ----------
    text: str
        The text captured

    likelihood: float
        The likelihood of the capture

    seconds: float
        The duration it took to do the processing

    site_id: str = "default"
        The id of the site where the text was captured

    session_id: Optional[str] = None
        The id of the session, if there is an active session

    wakeword_id: Optional[str] = None
        Optional id of wakeword used to activate ASR

    asr_tokens: Optional[List[List[AsrToken]]] = None
        Structured description of the tokens the ASR captured on for this intent

    lang: Optional[str] = None
        Language of the session
    """

    text: str
    likelihood: float
    seconds: float

    site_id: str = "default"
    session_id: typing.Optional[str] = None

    # ------------
    # Rhasspy only
    # ------------

    wakeword_id: typing.Optional[str] = None
    asr_tokens: typing.Optional[typing.List[typing.List[AsrToken]]] = None
    lang: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/asr/textCaptured"


# ----------------------------------------------------------------------------
# Rhasspy-only Messages
# ----------------------------------------------------------------------------


@dataclass
class AsrError(Message):
    """Error from ASR component.

    Attributes
    ----------
    site_id: str = "default"
        The id of the site where the error occurred

    error: str
        A description of the error that occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/error/asr"


@dataclass
class AsrTrain(Message):
    """Request to retrain ASR from intent graph.

    Attributes
    ----------
    graph_path: str
        Path to the graph file

    id: Optional[str] = None
        Unique id for training request

    graph_format: typing.Optional[str] = None
        Optional format of the graph file
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/asr/([^/]+)/train$")

    graph_path: str
    id: typing.Optional[str] = None
    graph_format: typing.Optional[str] = None
    sentences: typing.Optional[typing.Dict[str, str]] = None
    slots: typing.Optional[typing.Dict[str, typing.List[str]]] = None

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        site_id = kwargs.get("site_id", "+")
        return f"rhasspy/asr/{site_id}/train"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AsrTrain.TOPIC_PATTERN, topic) is not None

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AsrTrain.TOPIC_PATTERN, topic)
        assert match, "Not a train topic"
        return match.group(1)


@dataclass
class AsrTrainSuccess(Message):
    """Result from successful training.

    Attributes
    ----------

    id: Optional[str] = None
        Unique id from training request
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/asr/([^/]+)/trainSuccess$")

    id: typing.Optional[str] = None

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        site_id = kwargs.get("site_id", "+")
        return f"rhasspy/asr/{site_id}/trainSuccess"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AsrTrainSuccess.TOPIC_PATTERN, topic) is not None

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AsrTrainSuccess.TOPIC_PATTERN, topic)
        assert match, "Not a trainSuccess topic"
        return match.group(1)


@dataclass
class AsrAudioCaptured(Message):
    """Audio captured from ASR session.

    Attributes
    ----------
    wav_bytes: bytes
        Captured audio in WAV format
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/asr/([^/]+)/([^/]+)/audioCaptured$")

    wav_bytes: bytes

    def payload(self) -> typing.Union[str, bytes]:
        """Get binary/string for this message."""
        return self.wav_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """True if payload is not JSON."""
        return True

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def is_session_in_topic(cls) -> bool:
        """True if session id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        site_id = kwargs.get("site_id", "+")
        session_id = kwargs.get("site_id", "+")
        return f"rhasspy/asr/{site_id}/{session_id}/audioCaptured"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AsrAudioCaptured.TOPIC_PATTERN, topic) is not None

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AsrAudioCaptured.TOPIC_PATTERN, topic)
        assert match, "Not an audioCaptured topic"
        return match.group(1)

    @classmethod
    def get_session_id(cls, topic: str) -> typing.Optional[str]:
        """Get session id from a topic"""
        match = re.match(AsrAudioCaptured.TOPIC_PATTERN, topic)
        assert match, "Not an audioCaptured topic"
        return match.group(2)



"""Tests for rhasspyhermes.dialogue"""
from rhasspyhermes.dialogue import (
    DialogueContinueSession,
    DialogueEndSession,
    DialogueIntentNotRecognized,
    DialogueSessionEnded,
    DialogueSessionQueued,
    DialogueSessionStarted,
    DialogueStartSession,
)


def test_dialogue_continue_session():
    """Test DialogueContinueSession."""
    assert DialogueContinueSession.topic() == "hermes/dialogueManager/continueSession"


def test_dialogue_end_session():
    """Test DialogueEndSession."""
    assert DialogueEndSession.topic() == "hermes/dialogueManager/endSession"


def test_dialogue_intent_not_recognized():
    """Test DialogueIntentNotRecognized."""
    assert (
        DialogueIntentNotRecognized.topic()
        == "hermes/dialogueManager/intentNotRecognized"
    )


def test_dialogue_session_ended():
    """Test DialogueSessionEnded."""
    assert DialogueSessionEnded.topic() == "hermes/dialogueManager/sessionEnded"


def test_dialogue_session_queued():
    """Test DialogueSessionQueued."""
    assert DialogueSessionQueued.topic() == "hermes/dialogueManager/sessionQueued"


def test_dialogue_session_started():
    """Test DialogueSessionStarted."""
    assert DialogueSessionStarted.topic() == "hermes/dialogueManager/sessionStarted"


def test_dialogue_start_session():
    """Test DialogueStartSession."""
    assert DialogueStartSession.topic() == "hermes/dialogueManager/startSession"



"""Messages for text to speech."""
import typing
from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json

from .base import Message


@dataclass
class TtsSay(Message):
    """Send text to be spoken by the text to speech component.

    Note
    ----

    This is a low-level message. You should use the dialogue manager's
    :class:`rhasspyhermes.dialogue.DialogueStartSession` or
    :class:`rhasspyhermes.dialogue.DialogueContinueSession`
    messages.


    .. admonition:: MQTT message

      Topic
        ``hermes/tts/say``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - text
            - String
            - The text to be spoken.
          * - siteId
            - String
            - The id of the site where the text should be spoken. Defaults to ``"default"``.
          * - lang
            - String (optional)
            - The language code to use when saying the text.
          * - id
            - String (optional)
            - A request identifier. If provided, it will be passed back in the
              response message :class:`TtsSayFinished`.
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/tts/say' -m '{"text": "Ciao!", "lang": "it_IT"}'

    Example
    -------

    >>> from rhasspyhermes.tts import TtsSay
    >>> say = TtsSay(text="Ciao!", lang="it_IT")
    >>> say.topic()
    'hermes/tts/say'
    >>> say.payload()
    '{"text": "Ciao!", "siteId": "default", "lang": "it_IT", "id": null, "sessionId": null}'
    """

    text: str
    """The text to be spoken."""
    site_id: str = "default"
    """The id of the site where the text should be spoken."""
    lang: typing.Optional[str] = None
    """The language code to use when saying the text."""
    id: typing.Optional[str] = None
    """A request identifier. If provided, it will be passed back in the
    response message :class:`TtsSayFinished`."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/tts/say"``
        """
        return "hermes/tts/say"


@dataclass
class TtsSayFinished(Message):
    """Response published when the text to speech component has finished speaking.

    .. admonition:: MQTT message

      Topic
        ``hermes/tts/sayFinished``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where the text was spoken. Defaults to ``"default"``.
          * - id
            - String (optional)
            - Identifier from the request (:class:`TtsSay`).
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/tts/sayFinished'
    """

    site_id: str = "default"
    """The id of the site where the text was spoken. Defaults to ``"default"``."""
    id: typing.Optional[str] = None
    """Identifier from the request (:class:`TtsSay`)."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/tts/sayFinished"``
        """
        return "hermes/tts/sayFinished"


# -----------------------------------------------------------------------------
# Rhasspy Only
# -----------------------------------------------------------------------------


@dataclass
class GetVoices(Message):
    """Get the available voices for the text to speech system.

    Note
    ----
    This is a Rhasspy-only message.


    .. admonition:: MQTT message

      Topic
        ``hermes/tts/getVoices``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - id
            - String (optional)
            - Unique identifier passed to the response (:class:`Voices`).
          * - siteId
            - String
            - The id of the site to request voices from. Defaults to ``"default"``.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/tts/getVoices' -m '{"id": "abcd", "siteId": "default"}'

    Example
    -------
    >>> from rhasspyhermes.tts import GetVoices
    >>> g = GetVoices("abcd")
    >>> g.topic()
    'rhasspy/tts/getVoices'
    >>> g.payload()
    '{"id": "abcd", "siteId": "default"}'
    """

    id: typing.Optional[str] = None
    """Unique identifier passed to response (:class:`Voices`)."""
    site_id: str = "default"
    """Id of site to request voices from."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/tts/getVoices"``
        """
        return "rhasspy/tts/getVoices"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Voice:
    """Information about a single TTS voice."""

    voice_id: str
    """Unique identifier for voice."""
    description: typing.Optional[str] = None
    """Human-readable description of voice."""


@dataclass
class Voices(Message):
    """Response with the available voices for the text to speech system.
    This message is published in response to a :class:`GetVoices` request.

    Note
    ----
    This is a Rhasspy-only message.


    .. admonition:: MQTT message

      Topic
        ``hermes/tts/Voices``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - voices
            - List of JSON objects
            - List of available voices.
          * - id
            - String (optional)
            - Unique identifier from the request (:class:`GetVoices`).
          * - siteId
            - String
            - The id of the site where voices were requested. Defaults to ``"default"``.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/tts/voices'

    """

    voices: typing.List[Voice]
    """List of available voices."""
    id: typing.Optional[str] = None
    """Unique identifier from request (:class:`GetVoices`)."""
    site_id: str = "default"
    """Id of site where voices were requested."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/tts/voices"``
        """
        return "rhasspy/tts/voices"


@dataclass
class TtsError(Message):
    """This message is published by the text to speech system if an error has occurred.

    Note
    ----
    This is a Rhasspy-only message.


    .. admonition:: MQTT message

      Topic
        ``hermes/error/tts``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - error
            - String
            - A description of the error that occurred.
          * - siteId
            - String
            - Site where the error occurred. Defaults to ``"default"``.
          * - context
            - String (optional)
            - Additional information on the context in which the error occurred.
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/error/tts'

    Example
    -------

    >>> from rhasspyhermes.tts import TtsError
    >>> tts_error = TtsError(error="Unexpected error")
    >>> tts_error.topic()
    'hermes/error/tts'
    >>> tts_error.payload()
    '{"error": "Unexpected error", "siteId": "default", "context": null, "sessionId": null}'
    """

    error: str
    """A description of the error that occurred."""
    site_id: str = "default"
    """The id of the site where the error occurred."""
    context: typing.Optional[str] = None
    """Additional information on the context in which the error occurred."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/error/tts"``
        """
        return "hermes/error/tts"



"""Tests for rhasspyhermes.wake"""
from rhasspyhermes.wake import HotwordDetected, HotwordToggleOff, HotwordToggleOn

wakeword_id = "testWakeWord"


def test_hotword_detected():
    """Test HotwordDetected."""
    assert HotwordDetected.is_topic(HotwordDetected.topic(wakeword_id=wakeword_id))
    assert (
        HotwordDetected.get_wakeword_id(HotwordDetected.topic(wakeword_id=wakeword_id))
        == wakeword_id
    )


def test_hotword_toggle_on():
    """Test HotwordToggleOn."""
    assert HotwordToggleOn.topic() == "hermes/hotword/toggleOn"


def test_hotword_toggle_off():
    """Test HotwordToggleOff."""
    assert HotwordToggleOff.topic() == "hermes/hotword/toggleOff"



"""Messages for hermes/hotword"""
import re
import typing
from dataclasses import dataclass
from enum import Enum

from dataclasses_json import LetterCase, dataclass_json

from .base import Message


class HotwordToggleReason(str, Enum):
    """Reason for hotword toggle on/off.

    Values
    ------
    UNKNOWN
        Overrides all other reasons

    DIALOGUE_SESSION
        Dialogue session is active

    PLAY_AUDIO
        Audio is currently playing

    TTS_SAY
        Text to speech system is currently speaking
    """

    UNKNOWN = ""
    DIALOGUE_SESSION = "dialogueSession"
    PLAY_AUDIO = "playAudio"
    TTS_SAY = "ttsSay"


@dataclass
class HotwordToggleOn(Message):
    """Activate the wake word component.

    Attributes
    ----------
    site_id: str = "default"
        Id of the site where the hotword component should be enabled

    reason: HotwordToggleReason = UNKNOWN
        Reason for enabling the hotword component
    """

    site_id: str = "default"

    # ------------
    # Rhasspy only
    # ------------

    reason: HotwordToggleReason = HotwordToggleReason.UNKNOWN

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/hotword/toggleOn"


@dataclass
class HotwordToggleOff(Message):
    """Deactivate the wake word component.

    Attributes
    ----------
    site_id: str = "default"
        Id of the site where the hotword component should be disabled

    reason: HotwordToggleReason = UNKNOWN
        Reason for disabling the hotword component
    """

    site_id: str = "default"

    # ------------
    # Rhasspy only
    # ------------

    reason: HotwordToggleReason = HotwordToggleReason.UNKNOWN

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/hotword/toggleOff"


@dataclass
class HotwordDetected(Message):
    """Wake word component has detected a specific wake word.

    Attributes
    ----------
    model_id: str
        The id of the model that triggered the wake word

    model_version: str = ""
        The version of the model

    model_type: str = "personal"
        The type of the model. Possible values: universal or personal

    current_sensitivity: float = 1.0
        The sensitivity configured in the model at the time of the detection

    site_id: str = "default"
        The id of the site where the wake word was detected

    session_id: Optional[str] = None
        The desired id of the dialogue session created after detection.
        Leave empty to have one auto-generated.

    send_audio_captured: Optional[bool] = None
        True if audio captured from ASR should be emitted on
        rhasspy/asr/{site_id}/{session_id}/audioCaptured

    lang: Optional[str] = None
        Language of the detected wake word.
        Copied by dialogue manager into subsequent ASR, NLU messages.
    """

    TOPIC_PATTERN = re.compile(r"^hermes/hotword/([^/]+)/detected$")

    model_id: str
    model_version: str = ""
    model_type: str = "personal"
    current_sensitivity: float = 1.0
    site_id: str = "default"

    # ------------
    # Rhasspy only
    # ------------

    session_id: typing.Optional[str] = None
    send_audio_captured: typing.Optional[bool] = None
    lang: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message (wakeword_id)."""
        wakeword_id = kwargs.get("wakeword_id", "+")
        return f"hermes/hotword/{wakeword_id}/detected"

    @classmethod
    def get_wakeword_id(cls, topic: str) -> str:
        """Get wakeword_id from a topic"""
        match = re.match(HotwordDetected.TOPIC_PATTERN, topic)
        assert match, "Not a detected topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(HotwordDetected.TOPIC_PATTERN, topic) is not None


# -----------------------------------------------------------------------------
# Rhasspy Only Messages
# -----------------------------------------------------------------------------


@dataclass
class HotwordError(Message):
    """Error from Hotword component.

    Attributes
    ----------
    error: str
        A description of the error that occurred

    site_id: str = "default"
        The id of the site where the error occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic"""
        return "hermes/error/hotword"


@dataclass
class GetHotwords(Message):
    """Request to list available hotwords.

    Attributes
    ----------
    site_id: str = "default"
        Id of site where hotword component exists

    id: typing.Optional[str] = None
        Unique id passed to response
    """

    site_id: str = "default"
    id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic"""
        return "rhasspy/hotword/getHotwords"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Hotword:
    """Description of a single hotword.

    Attributes
    ----------
    model_id: str
        Unique ID of hotword model

    model_words: str
        Actual words used to activate hotword

    model_version: str = ""
        Model version

    model_type: str = "personal"
        Model type (personal, unversal)
    """

    model_id: str
    model_words: str
    model_version: str = ""
    model_type: str = "personal"


@dataclass
class Hotwords(Message):
    """Response to getHotwords.

    Attributes
    ----------
    models: typing.List[Hotword]
        List of available hotwords

    site_id: str = "default"
        Id of site where hotwords were requested

    id: typing.Optional[str] = None
        Unique id passed from request
    """

    models: typing.List[Hotword]
    site_id: str = "default"
    id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "rhasspy/hotword/hotwords"



"""Tests for rhasspyhermes.asr"""
from rhasspyhermes.asr import (
    AsrStartListening,
    AsrStopListening,
    AsrTextCaptured,
    AsrToggleOff,
    AsrToggleOn,
)


def test_asr_toggle_on():
    """Test AsrToggleOn."""
    assert AsrToggleOn.topic() == "hermes/asr/toggleOn"


def test_asr_toggle_off():
    """Test AsrToggleOff."""
    assert AsrToggleOff.topic() == "hermes/asr/toggleOff"


def test_asr_start_listening():
    """Test AsrStartListening."""
    assert AsrStartListening.topic() == "hermes/asr/startListening"


def test_asr_stop_listening():
    """Test AsrStopListening."""
    assert AsrStopListening.topic() == "hermes/asr/stopListening"


def test_asr_text_captured():
    """Test AsrTextCaptured."""
    assert AsrTextCaptured.topic() == "hermes/asr/textCaptured"



"""Messages for looking up/guessing word pronunciations."""
import typing
from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json

from rhasspyhermes.base import Message


@dataclass
class G2pPronounce(Message):
    """Get phonetic pronunciation for words.

    Attributes
    ----------
    id: Optional[str] = None
        Unique id for request

    site_id: str = "default"
        Id of site to request pronunciations from

    words: List[str]
        Words to guess pronunciations for

    session_id: typing.Optional[str] = None
        Id of active session, if there is one

    num_guesses: int = 5
        Maximum number of guesses to return for words not in dictionary
    """

    words: typing.List[str]
    id: typing.Optional[str] = None
    site_id: str = "default"
    session_id: typing.Optional[str] = None
    num_guesses: int = 5

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message."""
        return "rhasspy/g2p/pronounce"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class G2pPronunciation:
    """Phonetic pronunciation for a single word.

    Attributes
    ----------
    phonemes: List[str]
        Phonetic pronunciation for word

    guessed: bool = False
        True if this pronunciation was guessed using a g2p model.
        False if it came from a pronunciation dictionary.
    """

    phonemes: typing.List[str]
    guessed: typing.Optional[bool] = None


@dataclass
class G2pPhonemes(Message):
    """Response to g2p/pronounce.

    Attributes
    ----------
    id: Optional[str] = None
        Unique id from request

    site_id: str = "default"
        Id of site where pronunciations were requested

    word_phonemes: Dict[str, List[G2pPronunciation]]
        Guessed or looked up pronunciations

    session_id: typing.Optional[str] = None
        Id of active session, if there is one
    """

    word_phonemes: typing.Dict[str, typing.List[G2pPronunciation]]
    id: typing.Optional[str] = None
    site_id: str = "default"
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message."""
        return "rhasspy/g2p/phonemes"


@dataclass
class G2pError(Message):
    """Error from G2P component.

    Attributes
    ----------
    error: str
        A description of the error that occurred

    site_id: str = "default"
        The id of the site where the error occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message."""
        return "rhasspy/error/g2p"



"""Tests for rhasspyhermes.handle"""
from rhasspyhermes.handle import HandleToggleOff, HandleToggleOn


def test_handle_toggle_off():
    """Test HandleToggleOff."""
    assert HandleToggleOff.topic() == "rhasspy/handle/toggleOff"
    assert HandleToggleOff().payload() == '{"siteId": "default"}'


def test_handle_toggle_on():
    """Test HandleToggleOn."""
    assert HandleToggleOn.topic() == "rhasspy/handle/toggleOn"
    assert HandleToggleOn(site_id="satellite").payload() == '{"siteId": "satellite"}'



"""Utility methods for Rhasspy Hermes messages."""
import dataclasses
import typing


def only_fields(
    cls, message_dict: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    """Return a dict with only valid fields."""
    if dataclasses.is_dataclass(cls):
        field_names = set(f.name for f in dataclasses.fields(cls))
        valid_fields = set(message_dict.keys()).intersection(field_names)
        return {key: message_dict[key] for key in valid_fields}

    return message_dict



"""Intent and slot classes for NLU."""
import typing
from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Intent:
    """Intent object with a name and confidence score."""

    intent_name: str
    """The name of the detected intent."""
    confidence_score: float
    """The probability of the detection, between 0 and 1 (1 being sure)."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class SlotRange:
    """The range where a slot is found in the input text."""

    start: int
    """Index of the beginning (inclusive) of the slot value in the substituted input."""
    end: int
    """Index of the end (exclusive) of the slot value in the substituted input."""

    # ------------
    # Rhasspy only
    # ------------

    raw_start: typing.Optional[int] = None
    """Index of the beginning (inclusive) of the slot value in the unsubstituted input.

    Note
    ----

    This is a Rhasspy-only attribute."""
    raw_end: typing.Optional[int] = None
    """Index of the end (exclusive) of the slot value in the unsubstituted input.

    Note
    ----

    This is a Rhasspy-only attribute."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Slot:
    """Named entity in an intent slot."""

    entity: str
    """The entity of the slot."""
    value: typing.Dict[str, typing.Any]
    """The resolved value of the slot. Contains at least a ``"value"`` key."""
    slot_name: str = None  # type: ignore
    """The name of the slot."""
    raw_value: str = None  # type: ignore
    """The raw value of the slot, as it was in the input."""
    confidence: float = 0.0
    """Confidence score of the slot, between 0 and 1 (1 being confident)."""
    range: typing.Optional[SlotRange] = None
    """The range where the slot is found in the input text."""

    def __post_init__(self) -> None:
        """dataclasses post-init"""
        if self.slot_name is None:
            self.slot_name = self.entity

        if self.raw_value is None:
            self.raw_value = self.value.get("value")

    @property
    def start(self) -> int:
        """Get the start index (inclusive) of the slot value."""
        if self.range:
            return self.range.start

        return 0

    @property
    def raw_start(self) -> int:
        """Get the start index (inclusive) of the raw slot value."""
        value = None
        if self.range:
            value = self.range.raw_start

        if value is None:
            return self.start

        return value

    @property
    def end(self) -> int:
        """Get the end index (exclusive) of the slot value."""
        if self.range:
            return self.range.end

        return 1

    @property
    def raw_end(self) -> int:
        """Get the end index (exclusive) of the raw slot value."""
        value = None
        if self.range:
            value = self.range.raw_end

        if value is None:
            return self.end

        return value



"""Tests for rhasspyhermes.tts"""
from rhasspyhermes.tts import TtsSay, TtsSayFinished


def test_tts_say():
    """Test TtsSay."""
    assert TtsSay.topic() == "hermes/tts/say"


def test_tts_say_finished():
    """Test TtsSayFinished."""
    assert TtsSayFinished.topic() == "hermes/tts/sayFinished"



"""Tests for rhasspyhermes.nlu"""
from rhasspyhermes.nlu import NluError, NluIntent, NluIntentNotRecognized, NluQuery

intent_name = "testIntent"


def test_nlu_intent():
    """Test NluIntent."""
    assert NluIntent.is_topic(NluIntent.topic(intent_name=intent_name))
    assert (
        NluIntent.get_intent_name(NluIntent.topic(intent_name=intent_name))
        == intent_name
    )


def test_nlu_query():
    """Test NluQuery."""
    assert NluQuery.topic() == "hermes/nlu/query"


def test_nlu_intent_not_Recognized():
    """Test NluIntentNotRecognized."""
    assert NluIntentNotRecognized.topic() == "hermes/nlu/intentNotRecognized"


def test_nlu_error():
    """Test NluError."""
    assert NluError.topic() == "hermes/error/nlu"
